/** \file   dd_io_kf.c
 *  \brief  Implementation of IO functions for data files in Knuth format (for ZDDs)
 *  \author Takahisa Toda
 */
#include <string.h>

#include "my_def.h"
#include "my_hash.h"
#include "dd_wrapper.h"
#include "dd_io_kf.h"

static const int VAL_BASE_NUM = 10;
static const int INDEX_BASE_NUM  = 16;

static inline int cmp_krec(const void *a, const void*b);
static inline int parse_kf(char buf[], size_t *pf, size_t *pv, size_t *plo, size_t *phi);
static size_t zexport_kf_rec(my_zddp f, size_t *pid, struct htable *h, FILE *out);

static size_t recdepth = 0; //!< recursion depth

/**
 * \brief a record for a knuth form
 */
struct krec {
  size_t v;      //!< the label of a node
  size_t idf;    //!< the temporary index of a node
  size_t idlo;   //!< the temporary index of the LO child
  size_t idhi;   //!< the temporary index of the HI child
  my_zddp f;     //!< the index of a node, which is assigned by ZDD library.
};

static inline int cmp_krec(const void *a, const void*b)
{
  const struct krec c = *((const struct krec*)a);
  const struct krec d = *((const struct krec*)b);
  
  if (c.v < d.v)      return  1;
  else if (c.v > d.v) return -1;
  else                return  0;
}


/**  \brief Obtain information of a data file, and check consistency of file format.
 *  \param  pnrow        pointer to the number of rows (equivalently, the number of internal nodes)
 *  \param  pmaxval      pointer to the maximum value of a node label 
 *  \param  in           pointer to a dafa file
 *  \return MY_SUCCESS if successful; otherwise, MY_FAILURE.
 */
int getfileinfo_kf(size_t *pnrow, size_t *pmaxval, FILE *in)
{
  char buf[BUFSIZ]; size_t maxval = 0; size_t nrow = 0;
  fseek(in, 0L, SEEK_SET);
  while (fgets(buf, BUFSIZ, in) != NULL) {
    nrow++;
    // Find the string surrounded by '~' and '?'.
    int i = 0;
    while(buf[i] != '~' && buf[i] != '\0') i++;
    ENSURE_TRUE_MSG(buf[i] == '~', "invalid format\n");
    i++;
    int j = i;
    while(buf[j] != '?' && buf[j] != '\0') j++;
    ENSURE_TRUE_MSG(buf[j] == '?', "invalid format\n");
    buf[j] = '\0';
    size_t v = strtoull(buf+i, NULL, VAL_BASE_NUM);
    ENSURE_TRUE_MSG(v != 0, "invalid format\n");
    buf[j] = '?';
    if(maxval < v) maxval = v;
  }

  *pnrow = nrow; *pmaxval = maxval;
  return MY_SUCCESS;
}

/** \brief  Parse a single line in a Knuth form data file
 *  \param  buf  array of length BUFSIZ
 *  \param  pf   pointer to the index of a node
 *  \param  pv   pointer to a label
 *  \param  plo  pointer to the index of the LO child
 *  \param  phi  pointer to the index of the HI child
 *  \return MY_SUCCESS if successful; otherwise, MY_FAILURE.
 */
static inline int parse_kf(char buf[], size_t *pf, size_t *pv, size_t *plo, size_t *phi)
{
  int i = 0; int j = 0;

  // Find a node index
  while(buf[j] != ':' && buf[j] != '\0') j++;
  ENSURE_TRUE_MSG(buf[j] == ':', "invalid format\n");
  buf[j] = '\0';
  *pf = strtoull(buf, NULL, INDEX_BASE_NUM);
  j++;
  i = j;

  // Find the string surrounded by '~' and '?'.
  while(buf[i] != '~' && buf[i] != '\0') i++;
  ENSURE_TRUE_MSG(buf[i] == '~', "invalid format\n");
  i++;
  j = i;
  while(buf[j] != '?' && buf[j] != '\0') j++;
  ENSURE_TRUE_MSG(buf[j] == '?', "invalid format\n");
  buf[j] = '\0';
  *pv = strtoull(buf+i, NULL, VAL_BASE_NUM);
  ENSURE_TRUE_MSG(*pv != 0 && *pv < INFTY, "invalid value\n");
  j++;
  i = j;

  // Find the LO index
  while(buf[j] != ':' && buf[j] != '\0') j++;
  ENSURE_TRUE_MSG(buf[j] == ':', "invalid format\n");
  buf[j] = '\0';
  *plo = strtoull(buf+i, NULL, INDEX_BASE_NUM);
  j++;
  i = j;

  // Find the HI index
  while(buf[j] != ')' && buf[j] != '\0') j++;
  ENSURE_TRUE_MSG(buf[j] == ')', "invalid format\n");
  buf[j] = '\0';
  *phi = strtoull(buf+i, NULL, INDEX_BASE_NUM);
  j++;
  i = j;
 
  return MY_SUCCESS;
}

/**  \brief  Build a ZDD from input data in a Knuth format.
 *  \param  nrow        the number of rows (equivalently, the number of internal nodes)
 *  \param  in          pointer to a dafa file
 *  \return ZDD
 *  \note nrow must be greater than 0.
 */
my_zddp zimport_kf(size_t nrow, FILE *in)
{ 
  ENSURE_TRUE_MSG(nrow > 0, "there are no internal nodes");
  nrow += 2; // add the number of terminal nodes. (in this function, nrow means the number of all nodes, including terminal nodes.)
  struct krec *a = (struct krec*)malloc(sizeof(struct krec)*nrow);
  ENSURE_TRUE_MSG(a != NULL, "memory allocation failed\n");
  
  struct htable *h = ht_create(HASH_DEFAULT_SIZE);
  ENSURE_TRUE_MSG(h != NULL, "hash creation failed");

  // for the zdd_bot
  a[0].v = INFTY;
  a[0].idf  = 0;
  a[0].idlo = 0;
  a[0].idhi = 0;
  a[0].f = zdd_bot();

  // for the zdd_top
  a[1].v = INFTY;
  a[1].idf  = 1;
  a[1].idlo = 1;
  a[1].idhi = 1;
  a[1].f = zdd_top();

  // for internal nodes
  char buf[BUFSIZ];
  fseek(in, 0L, SEEK_SET);
  for (size_t i = 2; i < nrow && fgets(buf, BUFSIZ, in) != NULL; i++) {
    parse_kf(buf, &(a[i].idf), &(a[i].v), &(a[i].idlo), &(a[i].idhi));
    a[i].f = ZDD_NULL;
  }

  // Sort records in decreasing order with respect to node label
  qsort(a, nrow, sizeof(struct krec), cmp_krec);

  for(size_t i = 0; i < nrow; i++) {ht_add(a[i].idf, i, h);}
  for(size_t i = 2; i < nrow; i++) {
    size_t pos = 0;
    int res = ht_search(a[i].idlo, &pos, h); assert(res);
    my_zddp lo = a[pos].f;
    res = ht_search(a[i].idhi, &pos, h); assert(res);
    my_zddp hi = a[pos].f;
    a[i].f = zdd_unique(a[i].v, lo, hi);
    ENSURE_TRUE_MSG(a[i].f != ZDD_NULL, "ZDD operation failed");
  }

  my_zddp root = a[nrow-1].f;
  free(a); ht_destroy(h);
  return root;
}

/**  \brief  Export ZDD as a Knuth format file.
 *  \param  maxval the maximum value of a node label
 *  \param  f      a non-constant ZDD node 
 *  \param  out    pointer to a dafa file
 *  \return MY_SUCCESS if successful; otherwise, MY_FAILURE.
 *  \note an input ZDD must not be constant.
 */
int zexport_kf(size_t maxval, my_zddp f, FILE *out)
{
  ENSURE_TRUE_MSG(!zdd_isconst(f), "constant ZDD node is given");

  size_t index = 0;
  struct htable *h = ht_create(HASH_DEFAULT_SIZE);
  ENSURE_TRUE_MSG(h != NULL, "hash table creation failed");

  ht_add((size_t)zdd_bot(), index++, h);
  ht_add((size_t)zdd_top(), index++, h);

  assert(recdepth == 0);
  zexport_kf_rec(f, &index, h, out);
  assert(recdepth == 0);

  ht_destroy(h); 
  return MY_SUCCESS;
}


static size_t zexport_kf_rec(my_zddp f, size_t *pid, struct htable *h, FILE *out)
{
  int res; size_t t = 0;
  if((res = ht_search((size_t)f, &t, h)) == MY_SUCCESS) return t;

  INC_RECDEPTH(recdepth);  
  size_t lo = zexport_kf_rec(zdd_LO(f), pid, h, out);
  size_t hi = zexport_kf_rec(zdd_HI(f), pid, h, out);

  t = (*pid)++;
  fprintf(out, "%lx: (~%zu?%lx:%lx)\n", t, zdd_label(f), lo, hi);
  ht_add((size_t)f, t, h);
  DEC_RECDEPTH(recdepth);  
  return t;
}
