/** \file   my_io_setfam.c
 *  \brief  Implementation of IO functions for data files that represent set families. 
 *  \author Takahisa Toda
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "my_def.h"
#include "my_io_setfam.h"

/**  \brief Obtain information of a data file, and check consistency of file format.
 *  \param  pnrow        pointer to the number of rows, including empty rows.
 *  \param  pmaxval      pointer to the maximum value of an entry
 *  \param  ptotal_nent  pointer to the total number of entries in a datafile
 *  \param  in           pointer to a dafa file
 *  \return MY_SUCCESS if successful; otherwise, MY_FAILURE.
 *  \see in2setfam
 */
int getfileinfo(size_t *pnrow, size_t *pmaxval, size_t *ptotal_nent, FILE *in)
{
  fseek(in, 0L, SEEK_SET);
  char buf[MAX_LINESIZ];
  *pnrow = *pmaxval = *ptotal_nent = 0;
  while (fgets(buf, MAX_LINESIZ, in) != NULL) {
    char *sp = buf;
    while(*sp == SF_SEPARATOR) {assert(sp+1 < buf+MAX_LINESIZ); sp++;}
    char *ep = sp;
    size_t ne = 0;
    size_t tt = 0; // previous entry
    while(*sp != '\n' && *sp != '\0') {
      while(*ep != SF_SEPARATOR && *ep != '\n' && *ep != '\0') {
	assert(ep+1 < buf+MAX_LINESIZ);
	if(!isdigit(*ep)) {
	  fprintf(stderr, "error: invalid character at %zu line\n", *pnrow);
	  exit(EXIT_FAILURE);
	}
	ep++;
      }
      if(*ep == SF_SEPARATOR) {assert(ep < buf+MAX_LINESIZ); *ep = '\0'; ep++;}
      size_t t = strtoull(sp, NULL, 10); ne++;
      /* Check Input Begin*/
      if(t <= 0 || INFTY <= t) {
	fprintf(stderr, "error: invalid value, %s at %zu line\n", sp, *pnrow);
	exit(EXIT_FAILURE);
      }
      if(t <= tt) {
	fprintf(stderr, "error: not sorted, %zu\t%zu at %zu line\n", tt, t, *pnrow);
	exit(EXIT_FAILURE);
      }
      tt = t;
      /* Check Input End*/
      if(*pmaxval < t) *pmaxval = t;
      while(*ep == SF_SEPARATOR) {assert(ep+1 < buf+MAX_LINESIZ); ep++;}
      sp  = ep;
    }
    *ptotal_nent += ne;
    (*pnrow)++;
  }
  return MY_SUCCESS;
}


/**  \brief Read an input data file, and conver it to the data representation setfam.
 *  \param  nrow        the number of rows in a data file
 *  \param  total_nent  the total number of entries in a data file
 *  \param  in          pointer to a dafa file
 *  \note
 *  - sf->elem[i-1][j-1] is the j-th element in the i-th set. 
 *  - sf->elem[i-1][k] == INFTY if the i-th set contains exactly k elements.
 *  - the returned data structure must be destroyed when it becomes unnecessary.
 *  \see getfileinfo, destroy_setfam
 */
struct setfam *in2setfam(size_t nrow, size_t total_nent, FILE *in)
{
  fseek(in, 0L, SEEK_SET);
  char buf[MAX_LINESIZ];

  // Create setfam.
  const size_t memsize = total_nent + nrow;
  struct setfam *sf = (struct setfam*)malloc(sizeof(struct setfam));
  ENSURE_TRUE_MSG(sf != NULL, "memory allocation failed");
  ENSURE_TRUE(nrow != 0);
  sf->elem = (int**)malloc(sizeof(int*)*nrow);
  ENSURE_TRUE_MSG(sf->elem != NULL, "memory allocation failed");
  ENSURE_TRUE(memsize != 0);
  sf->mem = (int*)malloc(sizeof(int)*memsize);
  ENSURE_TRUE_MSG(sf->mem != NULL, "memory allocation failed");  
  sf->nset = nrow;
  sf->memsize = memsize;

  // Convert to setfam.
  int **ee = sf->elem; int *mm = sf->mem;
  while (fgets(buf, MAX_LINESIZ, in) != NULL) {
    assert(ee < sf->elem + nrow);
    *ee = mm; ee++;
    char *sp = buf;
    while(*sp == SF_SEPARATOR) sp++;
    char *ep = sp;
    while(*sp != '\n' && *sp != '\0') {
      while(*ep != SF_SEPARATOR && *ep != '\n' && *ep != '\0') ep++;
      if(*ep == SF_SEPARATOR) {*ep = '\0'; ep++;}
      assert(mm < sf->mem + memsize);
      *mm = (int)strtoull(sp, NULL, 10); mm++;
      while(*ep == SF_SEPARATOR) ep++;
      sp  = ep;
    }
    assert(mm < sf->mem + memsize);
    *mm = INFTY; mm++; 
  }
  return sf;
}

/* \brief Free all memory areas used.
 * \param sf pointer to a setfam
 * \see in2setfam
 */
void destroy_setfam(struct setfam *sf)
{
  assert(sf != NULL);
  free(sf->elem); sf->elem = NULL;
  free(sf->mem);  sf->mem = NULL;
  free(sf);       sf = NULL;
}


